/*
 *	check/cmemcpy.c: part of libhmem(high speed memory operation library)
 *	Copyright  1994,95  씎Y(oo), HGG00736@niftyserve.or.jp
 */
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"hmemory.h"

#define	MEMCPY_BUFSIZ	0x400
#define	memclr(p,n)		memset((p),0,(n))

char*	dp;
char*	sp;
char buf0[MEMCPY_BUFSIZ];
char buf1[MEMCPY_BUFSIZ];
char buf2[MEMCPY_BUFSIZ];

#if	0
int	dumpstack(int d0, int d1, int d2, int a0, int a1, int a2)
{
	printf("d0=%p, d1=%p, d2=%p, a0=%p, a1=%p, a2=%p\n", d0, d1, d2, a0, a1, a2);
}
#endif

static void set_buf(char*p)
{
	int	i, n;

	for (n = 1, i = 0; i < MEMCPY_BUFSIZ; i++) {
		p[i] = n++;
		if ((n &= 0xff) == 0) n++;
	}
}

static void	dump_buf(unsigned char* p, int n)
{
	int	i;

	printf("%p\n", p);
	for (i = 0; i < n; i++) {
		printf("%02x", p[i]);
		if ((i&0x1f) == 31) fputc('\n', stdout);
		else if ((i&0x7) == 7) fputc('-', stdout);
		else fputc(' ', stdout);
	}
}

static void	check_buf(char* p1, char* p2, int n)
{
	int	dofs, sofs;
	char*	p;

	dofs = p1 - dp;
	sofs = p2 - sp;
	printf("dp=%p, dofs=%d, sp=%p, sofs=%d, n=%d\n", dp, dofs, sp, sofs, n);
	memclr(dp, MEMCPY_BUFSIZ);
	p = memcpy(p1, p2, n);
	if (p != p1
	||  memcmp(p1, p2, n)
	||  memcmp(sp, buf0, MEMCPY_BUFSIZ)
	||  (dofs && _memcchr(dp, 0, dofs))
	||  _memcchr(p1+n, 0, MEMCPY_BUFSIZ - n - dofs)) {
		if (p != p1) printf("p != p1\n");
		if (memcmp(p1, p2, n)) printf("memcmp(p1, p2, n)\n");
		if (memcmp(sp, buf0, MEMCPY_BUFSIZ)) printf("memcmp(sp, buf0, MEMCPY_BUFSIZ)\n");
		if (dofs && _memcchr(dp, 0, dofs)) printf("(dofs && _memcchr(dp, 0, dofs))\n");
		if (_memcchr(p1+n, 0, MEMCPY_BUFSIZ - n - dofs)) printf("_memcchr(p1+n, 0, MEMCPY_BUFSIZ - n - dofs))\n");
		fprintf(stderr, "difference %p - %p, %d\n", p1, p2, n);
		dump_buf((unsigned char*)sp, MEMCPY_BUFSIZ);
		dump_buf((unsigned char*)dp, MEMCPY_BUFSIZ);
		exit(1);
	}
#if	0
	dump_buf(sp, MEMCPY_BUFSIZ);
	dump_buf(dp, MEMCPY_BUFSIZ);
	getch();
#endif
}

static void	check(char* p1, char* p2, int n)
{
	check_buf(p1+0, p2+0, n);
	check_buf(p1+1, p2+0, n);
	check_buf(p1+2, p2+0, n);
	check_buf(p1+3, p2+0, n);

	check_buf(p1+0, p2+1, n);
	check_buf(p1+1, p2+1, n);
	check_buf(p1+2, p2+1, n);
	check_buf(p1+3, p2+1, n);

	check_buf(p1+0, p2+2, n);
	check_buf(p1+1, p2+2, n);
	check_buf(p1+2, p2+2, n);
	check_buf(p1+3, p2+2, n);

	check_buf(p1+0, p2+3, n);
	check_buf(p1+1, p2+3, n);
	check_buf(p1+2, p2+3, n);
	check_buf(p1+3, p2+3, n);
}

void	main(void)
{
	set_buf(buf0);
	set_buf(buf1);
	sp = buf1;
	dp = buf2;
	check(dp, sp, 1);
	check(dp, sp, 4);
	check(dp, sp, 8);
	check(dp, sp, 16);
	check(dp, sp, 32);
	check(dp, sp, 256);
	check(dp, sp, 512);

	set_buf(buf2);
	dp = buf1;
	sp = buf2;
	check(dp, sp, 1);
	check(dp, sp, 4);
	check(dp, sp, 8);
	check(dp, sp, 16);
	check(dp, sp, 32);
	check(dp, sp, 256);
	check(dp, sp, 512);

	printf("ok\n");
	exit(0);
}
