*
*  memcpy.s: part of libhmem(high speed memory operation library)
*  Copyright  1992-95  씎Y(oo), HGG00736@niftyserve.or.jp
*
.ifndef	MPU
MPU	equ	68000
.endif
	xref	__hmem_version

count1	equ	16		* limit count of longword copy(count1 > 12)
count2	equ	256		* limit count of block copy(count2 < 0x10000)

*
*	stack frame definitions
*
	.offset	4		* for memcpy(), memmove()
dst:	ds.l	1
src:	ds.l	1
len:	ds.l	1

	.offset	4		* for movedata(), movmem(), bcopy()
sptr	ds.l	1
dptr	ds.l	1

*
*	register defintions
*
~dst	.reg	a0
~src	.reg	a1
~len0	.reg	a2
~dst0	.reg	d0
~lenH	.reg	d1
~len	.reg	d2

~list	.reg	d3-d7/a3-a5
regsiz	.equ	32

	.text
	.even
.if	MPU>=68020
.ifdef	QUAD_ALIGN
	.quad
.endif
.endif
**
** void* movedata(const void* src, void* dst, size_t len)
**
	xdef	_movedata
_movedata:
**
** void* movmem(const void* src, void* dst, size_t len)
**
	xdef	_movmem
_movmem:
**
** void* bcopy(const void* src, void* dst, size_t len)
**
	xdef	_bcopy
_bcopy:
	move.l	dptr(sp),~dst
	move.l	sptr(sp),~src
	move.l	len(sp),a2
	bra.s	memcpy

	xdef	_hmemcpy
_hmemcpy:
**
** void* memmove(void* dst, const void* src, size_t len)
**
	xdef	_memmove
_memmove:
**
** void* memcpy(void* dst, const void* src, size_t len)
**
	xdef	_memcpy
_memcpy:
	movem.l dst(sp),~dst/~src/~len0	* a0=dst, a1=src, a2=len
memcpy:
	move.l	~dst,~dst0		* a0=dst, a1=src, a2=len, d0=dst
*	beq.s	ncopy_byte_end		* dst == NULL
	move.l	~len0,~len		* a0=dst, a1=src, a2=len, d0=dst, d2=len
	beq.s	ncopy_byte_end		* 
	cmpa.l	~dst,~src		* check copy direction
	beq.s	ncopy_byte_end		* the same area
	bcs	reverse_copy		* ->reverse copy

normal_copy:
	cmpi.l	#count1,~len		* if (~len < count1)
	bhs.s	ncopy_check_align	* then ncopy_check_align

ncopy_byte:
	subq.w	#1,~len			* adjust len

ncopy_byte_loop:
	move.b	(~src)+,(~dst)+
	dbra	~len,ncopy_byte_loop
ncopy_byte_end
	move.l	dst(sp),d0
	rts

ncopy_byte8:
	move.w	~len,~len0
	lsr.l	#3,~len			* len /= 8
	subq.l	#1,~len
	move.l	~len,~lenH
	swap	~lenH

ncopy_byte8_loop:			* thanks to Make up
	move.b	(~src)+,(~dst)+
	move.b	(~src)+,(~dst)+
	move.b	(~src)+,(~dst)+
	move.b	(~src)+,(~dst)+		*  4
	move.b	(~src)+,(~dst)+
	move.b	(~src)+,(~dst)+
	move.b	(~src)+,(~dst)+
	move.b	(~src)+,(~dst)+		*  8
	dbra	~len,ncopy_byte8_loop
	dbra	~lenH,ncopy_byte8_loop
	bra.s	ncopy_long_rest

ncopy_check_align:
*	 d0=dst, d1=src, d2=len, a0=dst, a1=src, a2=len
	move.w	~src,d1	
	eor.w	~dst0,d1		* d1= src^dst
	btst.l	#0,d1			* if ((src^dst) & 1)
	bne.s	ncopy_byte8		* then ncopy_byte4

	btst.l	#0,~dst0		* odd address?
	beq.s	ncopy_align_even	* if even then skip

	move.b	(~src)+,(~dst)+
	subq.l	#1,~len
*	beq.s	ncopy_word_end
ncopy_align_even:
.if	MPU>=68040
	btst.l	#1,d1			* if ((src^dst) & 2)
	beq.s	ncopy_check_align2

ncopy_word8:
	move.l	~len,~len0
	lsr.l	#3,~len			* len /= 8
	subq.l	#1,~len
	move.l	~len,~lenH
	swap	~lenH

ncopy_word8_loop:
	move.w	(~src)+,(~dst)+
	move.w	(~src)+,(~dst)+
	move.w	(~src)+,(~dst)+
	move.w	(~src)+,(~dst)+		* 8
	dbra	~len,ncopy_word8_loop
	dbra	~lenH,ncopy_word8_loop
	bra.s	ncopy_long_rest
.endif

ncopy_check_align2:
.if	MPU>=68020
	btst.l	#1,~dst0		* even address
	beq.s	ncopy_align_quad

	move.w	(~src)+,(~dst)+
	subq.l	#2,~len
*	beq.s	ncopy_word_end
.endif
ncopy_align_quad:
	move.l	~len,~len0
	cmp.l	#count2,~len
	bcc.s	ncopy_block

ncopy_long:
	lsr.l	#3,~len			* len /= 8
	subq.l	#1,~len
*	move.l	~len,~lenH		* count2 < 0x10000
*	swap	~lenH

ncopy_long8_loop:
	move.l	(~src)+,(~dst)+
	move.l	(~src)+,(~dst)+
	dbra	~len,ncopy_long8_loop
*	dbra	~lenH,ncopy_long8_loop
ncopy_long_rest:
	move.w	~len0,~len
	and.w	#7,~len
	bne	ncopy_byte
ncopy_long_end:
	rts

ncopy_block:
	movem.l	~list,-(sp)
	lsr.l	#7,d2			* len /= 128
	subq.l	#1,~len
	move.l	~len,~lenH
	swap	~lenH

ncopy_block_loop:
	movem.l	regsiz*0(~src),~list
	movem.l	~list,regsiz*0(~dst)	*  32
	movem.l	regsiz*1(~src),~list
	movem.l	~list,regsiz*1(~dst)	*  64
	movem.l	regsiz*2(~src),~list
	movem.l	~list,regsiz*2(~dst)	*  96
	movem.l	regsiz*3(~src),~list
	movem.l	~list,regsiz*3(~dst)	* 128
	lea.l	128(~src),~src
	lea.l	128(~dst),~dst
	dbra	~len,ncopy_block_loop
	dbra	~lenH,ncopy_block_loop

	movem.l	(sp)+,~list
	move.w	~len0,~len
	andi.w	#127,~len
	beq.s	ncopy_long_end
	cmpi.w	#count1,~len
	bhs.s	ncopy_long
	bra	ncopy_byte

reverse_copy:
	adda.l	~len,~dst		* dst += len
	adda.l	~len,~src		* src += len
	cmpi.l	#count1,~len		. count1oCg̎̓oCg]
	bhs.s	rcopy_check_align

rcopy_byte:
	subq.w	#1,~len			* adjust length

rcopy_byte_loop:
	move.b	-(~src),-(~dst)
	dbra	~len,rcopy_byte_loop
	move.l	~dst,d0
rcopy_byte_end:
	rts

rcopy_byte8:
	move.w	~len,~len0
	lsr.l	#3,~len			* len /= 8
	subq.l	#1,~len
	move.l	~len,~lenH
	swap	~lenH

rcopy_byte8_loop:			* thanks to Make up
	move.b	-(~src),-(~dst)
	move.b	-(~src),-(~dst)
	move.b	-(~src),-(~dst)
	move.b	-(~src),-(~dst)		*  4
	move.b	-(~src),-(~dst)
	move.b	-(~src),-(~dst)
	move.b	-(~src),-(~dst)
	move.b	-(~src),-(~dst)		*  8
	dbra	~len,rcopy_byte8_loop
	dbra	~lenH,rcopy_byte8_loop
	bra.s	rcopy_long_rest
*	move.w	~len0,~len
*	andi.w	#7,~len
*	bne.s	rcopy_byte
*	move.l	~dst,d0
*	rts

rcopy_check_align:
*	d0=dst+len, d1=src+len, d2=len, a0=dst+len, a1=src+len, a2=len
	move.w	~dst,~dst0
	move.w	~src,d1

	eor.w	~dst0,d1		* (dptr^sptr)
	btst.l	#0,d1			* if ((dptr^sptr) & 1)
	bne.s	rcopy_byte8		* then rcopy_byte8

	btst.l	#0,~dst0		* odd address
	beq.s	rcopy_align_even

	move.b	-(~src),-(~dst)
	subq.l	#1,~len
*	beq.s	rcopy_word_end
rcopy_align_even:
.if	MPU>=68040
	btst.l	#1,d1			* if ((src^dst) & 2)
	beq.s	rcopy_check_align2

rcopy_word8:
	move.l	~len,~len0
	lsr.l	#3,~len			* len /= 8
	subq.l	#1,~len
	move.l	~len,~lenH
	swap	~lenH

rcopy_word8_loop:
	move.w	-(~src),-(~dst)
	move.w	-(~src),-(~dst)
	move.w	-(~src),-(~dst)
	move.w	-(~src),-(~dst)		* 8
	dbra	~len,rcopy_word8_loop
	dbra	~lenH,rcopy_word8_loop
	bra.s	rcopy_long_rest
*	move.w	~len0,~len
*	andi.w	#7,~len
*	bne.s	rcopy_byte
*	move.l	~dst,d0
*	rts
.endif

rcopy_check_align2:
.if	MPU>=68020
	btst.l	#1,d0			* even address
	beq.s	rcopy_align_quad

	move.w	-(~src),-(~dst)
	subq.l	#2,~len
*	beq.s	rcopy_word_end
.endif
rcopy_align_quad:
	move.l	~len,~len0
	cmp.l	#count2,~len
	bcc.s	rcopy_block

rcopy_long:
	lsr.l	#3,~len			* len /= 8
	subq.l	#1,~len
*	move.l	d2,d1			* count2 < 0x10000
*	swap	d1

rcopy_long8_loop:
	move.l	-(~src),-(~dst)
	move.l	-(~src),-(~dst)		* 8
	dbra	~len,rcopy_long8_loop
*	dbra	~lenH,rcopy_long8_loop
rcopy_long_rest:
	move.w	~len0,~len
	andi.w	#7,~len
	bne	rcopy_byte
rcopy_long_end:
	move.l	~dst,d0
	rts

rcopy_block:
	movem.l	~list,-(sp)
	lsr.l	#7,~len			* len /= 128
	subq.l	#1,~len
	move.l	~len,~lenH
	swap	~lenH

rcopy_block_loop:
	movem.l	-regsiz*1(~src),~list
	movem.l	~list,-regsiz*1(~dst)	*  32
	movem.l	-regsiz*2(~src),~list
	movem.l	~list,-regsiz*2(~dst)	*  64
	movem.l	-regsiz*3(~src),~list
	movem.l	~list,-regsiz*3(~dst)	*  96
	movem.l	-regsiz*4(~src),~list
	movem.l	~list,-regsiz*4(~dst)	* 128
	lea.l	-128(~src),~src
	lea.l	-128(~dst),~dst
	dbra	~len,rcopy_block_loop
	dbra	~lenH,rcopy_block_loop

	movem.l	(sp)+,~list
	move.w	~len0,~len
	andi.w	#127,~len
	beq.s	rcopy_long_end
	cmpi.w	#count1,~len
	bhs.s	rcopy_long
	bra	rcopy_byte

	.end
