/*
 *	test/tmemcpy.c: part of libhmem(high speed memory operation library)
 *	Copyright  1994,95  씎Y(oo), HGG00736@niftyserve.or.jp
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>
#ifdef	HAVE_HMEMORY_H
#include	"hmemory.h"
#endif

static long	loop_count;
static char*	dst = NULL;
static char*	src = NULL;

static void	disptime(char* msg, clock_t count)
{
	float	t;

	t = (float)count * 10000.0 / (float)loop_count;	/* us */
	printf(", %s: ", msg);
	if (t >= 1000000.0) {
		t /= 1000000.0;
		t += 0.005;
		printf("%6.2f(s)", t);
	}
	else if (t >= 1000.0) {
		t /= 1000.0;
		t += 0.005;
		printf("%6.2f(ms)", t);
	}
	else {
		t += 0.005;
		printf("%6.2f(us)", t);
	}
}


static void	sub(int size)
{
	int	i;
	clock_t	start, stop, loop_time;
#ifdef	HAVE_HMEMORY_H
	int	cnt;
#else
	char*	sp;
	char*	dp;
#endif

	start = clock();
	for (i = 0; i < loop_count; i++) {
 		;
	}
	stop = clock();
	loop_time = stop - start;

#ifndef	HAVE_HMEMORY_H
	dp = dst;
	sp = src;
	start = clock();
	for (i = 0; i < loop_count; i++) {
		memcpy(dp, sp, size);
	}
	stop = clock();
	disptime("+0", stop - start - loop_time);

	sp++;
	start = clock();
	for (i = 0; i < loop_count; i++) {
		memcpy(dp, sp, size);
	}
	stop = clock();
	disptime("+1", stop - start - loop_time);

	sp++;
	start = clock();
	for (i = 0; i < loop_count; i++) {
		memcpy(dp, sp, size);
	}
	stop = clock();
	disptime("+2", stop - start - loop_time);

	sp++;
	start = clock();
	for (i = 0; i < loop_count; i++) {
		memcpy(dp, sp, size);
	}
	stop = clock();
	disptime("+3", stop - start - loop_time);

#else
	if (size >= sizeof(short)) {
		cnt = size/sizeof(short);
		start = clock();
		for (i = 0; i < loop_count; i++) {
			_memcpyw((short*)dst, (CONST short*)src, cnt);
		}
		stop = clock();
		disptime("W", stop - start - loop_time);
	}

	if (size >= sizeof(long)) {
		cnt = size/sizeof(long);
		start = clock();
		for (i = 0; i < loop_count; i++) {
			_memcpyl((long*)dst, (CONST long*)src, cnt);
		}
		stop = clock();
		disptime("L", stop - start - loop_time);
	}

	if (size >= 256) {
		cnt = size/256;
		start = clock();
		for (i = 0; i < loop_count; i++) {
			_blkcpy((long*)dst, (CONST long*)src, cnt);
		}
		stop = clock();
		disptime("B", stop - start - loop_time);
	}
#endif
}


void	main(void)
{
	int	size;

	if ((dst = malloc(0x10000+16)) == NULL) exit(1);
	memset(dst, 0, 0x10000+16);
	if ((src = malloc(0x10000+16)) == NULL) exit(1);
	memset(src, 0, 0x10000+16);

	for (loop_count = 655360, size = 1; size <= 0x10000; size <<= 1, loop_count >>= 1) {
		printf("size=%5d, loop=%6ld", size, loop_count);
		sub(size);
		printf("\n", size, loop_count);
	}

	free(src);
	free(dst);
	exit(0);
}
